import	os
import	unittest
from	pygwarts.irma.shelve	import LibraryShelf
from	pygwarts.irma.contrib	import LibraryContrib
from	pygwarts.tests.irma		import IrmaTestCase








class ItemsTest(IrmaTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.ITEMS_HANDLER): os.remove(cls.ITEMS_HANDLER)

	@classmethod
	def setUpClass(cls):
		class ItemsTestCase(LibraryShelf):
			class loggy(LibraryContrib):

				init_name	= "ItemsTest"
				init_level	= 10
				handler		= cls.ITEMS_HANDLER

		cls.make_loggy_file(cls, cls.ITEMS_HANDLER)
		cls.test_case = ItemsTestCase()

	def setUp(self) : self.test_case.unload(magical=True)
	def test_is_mapping_check(self):

		self.assertTrue(hasattr(self.test_case, "keys"))
		self.assertTrue(hasattr(self.test_case, "__getitem__"))




	def test_real_setitem(self):

		for N,item in enumerate(( "LOL", 1, .1, 0, ( "LOL", ), None ),1):
			with self.subTest(item=item, N=N):

				self.test_case[item] = "KEK"
				self.assertEqual(len(self.test_case.real_shelf), N)
				self.assertEqual(len(self.test_case.magical_shelf), N)
				self.assertEqual(len(self.test_case.real_shelf._locker_), N)
				self.assertEqual(len(self.test_case.magical_shelf._locker_), N)
				self.assertEqual(self.test_case.diff,0)
				self.assertTrue(self.test_case.modified)


		self.assertEqual(self.test_case.keys(), [ "LOL", 1, .1, 0, ( "LOL", ), None ])
		self.assertEqual(self.test_case.keysof("KEK"), [ "LOL", 1, .1, 0, ( "LOL", ), None ])
		self.assertEqual(self.test_case.keysof("CHEBUREK"), [])




	def test_real_invalid_setitem(self):

		for item in [ "LOL","KEK" ], { "LOL": "KEK" }, { "LOL","KEK" }:
			with self.subTest(item=item):

				self.test_case[item] = "CHEBUREK"
				self.assertEqual(len(self.test_case.real_shelf), 0)
				self.assertEqual(len(self.test_case.magical_shelf), 0)
				self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
				self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
				self.assertEqual(self.test_case.diff,0)
				self.assertFalse(self.test_case.modified)


		self.assertEqual(self.test_case.keys(), [])
		self.assertEqual(self.test_case.keysof("CHEBUREK"), [])




	def test_real_setitem_edge_case(self):

		self.test_case[None] = None
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case.keys(), [ None ])
		self.assertEqual(self.test_case.keysof(None), [ None ])


		# Imitation of situation, when Shelf is grabbed and modified
		self.test_case.modified = False
		self.test_case[None] = None
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertFalse(self.test_case.modified)
		self.assertEqual(self.test_case.keys(), [ None ])
		self.assertEqual(self.test_case.keysof(None), [ None ])


		self.test_case.modified = False
		self.test_case[None] = "LOL-KEK-CHEBUREK"
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case.keys(), [ None ])
		self.assertEqual(self.test_case.keysof("LOL-KEK-CHEBUREK"), [ None ])




	def test_real_replace_item(self):

		self.test_case["LOL"] = "KEK"
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case.keys(), [ "LOL" ])
		self.assertEqual(self.test_case.keysof("CHEBUREK"), [])
		self.assertEqual(self.test_case.keysof("KEK"), [ "LOL" ])


		# Imitation of situation, when Shelf is grabbed and modified
		self.test_case.modified = False
		self.assertFalse(self.test_case.modified)
		self.test_case["LOL"] = "CHEBUREK"
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case.keys(), [ "LOL" ])
		self.assertEqual(self.test_case.keysof("KEK"), [])
		self.assertEqual(self.test_case.keysof("CHEBUREK"), [ "LOL" ])


		self.test_case.modified = False
		self.assertFalse(self.test_case.modified)
		self.test_case["LOL"] = "CHEBUREK"
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertFalse(self.test_case.modified)
		self.assertEqual(self.test_case.keys(), [ "LOL" ])
		self.assertEqual(self.test_case.keysof("KEK"), [])
		self.assertEqual(self.test_case.keysof("CHEBUREK"), [ "LOL" ])




	def test_real_getitem(self):

		self.test_case["LOL"] = "KEK"
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)


		self.assertEqual(self.test_case["LOL"], "KEK")
		self.assertEqual(self.test_case["KEK"], None)
		self.assertNotEqual(self.test_case["LOL"], "CHEBUREK")
		self.assertNotEqual(self.test_case["KEK"], "LOL")


		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)




	def test_real_contains_item(self):

		self.test_case["LOL"] = "KEK"
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)


		self.assertTrue("LOL" in self.test_case)
		self.assertFalse("KEK" in self.test_case)
		self.assertFalse([ "KEK" ] in self.test_case)




	def test_real_delitem(self):

		self.test_case["LOL"] = "KEK"
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case.keys(), [ "LOL" ])
		self.assertEqual(self.test_case.keysof("KEK"), [ "LOL" ])
		self.assertEqual(self.test_case.keysof("CHEBUREK"), [])


		with self.assertLogs("ItemsTest", 10) as case_loggy:

			del self.test_case["LOL"]
			self.assertEqual(len(self.test_case.real_shelf), 0)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,0)
			del self.test_case["KEK"]
			del self.test_case[["KEK"]]


		self.assertEqual(self.test_case.keys(), [])
		self.assertEqual(self.test_case.keysof("KEK"), [])
		self.assertEqual(self.test_case.keysof("CHEBUREK"), [])
		self.assertIn("DEBUG:ItemsTest:Removed \"LOL\", \"KEK\" pair", case_loggy.output)
		self.assertIn(

			f"DEBUG:ItemsTest:Key \"KEK\" not in {self.test_case.real_shelf} KeyChest",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:ItemsTest:Key \"KEK\" not in {self.test_case.magical_shelf} KeyChest",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:ItemsTest:Impossible key \"[\'KEK\']\" for {self.test_case.real_shelf} KeyChest",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:ItemsTest:Impossible key \"[\'KEK\']\" for {self.test_case.magical_shelf} KeyChest",
			case_loggy.output
		)








	def test_magical_setitem(self):

		for N,item in enumerate(( "LOL", 1, .1, 0, ( "LOL", ), None ),1):
			with self.subTest(item=item, N=N):

				self.test_case(item, "KEK")
				self.assertEqual(len(self.test_case.real_shelf), 0)
				self.assertEqual(len(self.test_case.magical_shelf), N)
				self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
				self.assertEqual(len(self.test_case.magical_shelf._locker_), N)
				self.assertEqual(self.test_case.diff,-N)
				self.assertTrue(self.test_case.modified)


		self.assertEqual(self.test_case().keys(), [ "LOL", 1, .1, 0, ( "LOL", ), None ])
		self.assertEqual(self.test_case().keysof("KEK"), [ "LOL", 1, .1, 0, ( "LOL", ), None ])
		self.assertEqual(self.test_case().keysof("CHEBUREK"), [])




	def test_magical_invalid_setitem(self):

		for item in [ "LOL","KEK" ], { "LOL": "KEK" }, { "LOL","KEK" }:
			with self.subTest(item=item):

				self.test_case(item, "CHEBUREK")
				self.assertEqual(len(self.test_case.real_shelf), 0)
				self.assertEqual(len(self.test_case.magical_shelf), 0)
				self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
				self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
				self.assertEqual(self.test_case.diff,0)
				self.assertFalse(self.test_case.modified)


		self.assertEqual(self.test_case().keys(), [])
		self.assertEqual(self.test_case().keysof("CHEBUREK"), [])




	def test_magical_setitem_edge_case(self):

		self.test_case(None,None)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case().keys(), [ None ])
		self.assertEqual(self.test_case().keysof(None), [ None ])


		# Imitation of situation, when Shelf is grabbed and modified
		self.test_case.modified = False
		self.test_case(None,None)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)
		self.assertEqual(self.test_case().keys(), [ None ])
		self.assertEqual(self.test_case().keysof(None), [ None ])


		self.test_case.modified = False
		self.test_case(None,"LOL-KEK-CHEBUREK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case().keys(), [ None ])
		self.assertEqual(self.test_case().keysof("LOL-KEK-CHEBUREK"), [ None ])




	def test_magical_replace_item(self):

		self.test_case("LOL","KEK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case().keys(), [ "LOL" ])
		self.assertEqual(self.test_case().keysof("CHEBUREK"), [])
		self.assertEqual(self.test_case().keysof("KEK"), [ "LOL" ])


		# Imitation of situation, when Shelf is grabbed and modified
		self.test_case.modified = False
		self.assertFalse(self.test_case.modified)
		self.test_case("LOL","CHEBUREK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case().keys(), [ "LOL" ])
		self.assertEqual(self.test_case().keysof("KEK"), [])
		self.assertEqual(self.test_case().keysof("CHEBUREK"), [ "LOL" ])


		self.test_case.modified = False
		self.assertFalse(self.test_case.modified)
		self.test_case("LOL", "CHEBUREK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)
		self.assertEqual(self.test_case().keys(), [ "LOL" ])
		self.assertEqual(self.test_case().keysof("KEK"), [])
		self.assertEqual(self.test_case().keysof("CHEBUREK"), [ "LOL" ])




	def test_magical_getitem(self):

		self.test_case("LOL","KEK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)

		with self.assertNoLogs("ItemsTest", 30) as tcloggy1:
			self.assertEqual(self.test_case("LOL"), "KEK")

		with self.assertNoLogs("ItemsTest", 30) as tcloggy2:
			self.assertEqual(self.test_case("KEK"), None)

		with self.assertNoLogs("ItemsTest", 30) as tcloggy3:
			self.assertNotEqual(self.test_case("LOL"), "CHEBUREK")

		with self.assertNoLogs("ItemsTest", 30) as tcloggy4:
			self.assertNotEqual(self.test_case("KEK"), "LOL")


		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)




	def test_magical_contains_item(self):

		self.test_case("LOL","KEK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)

		self.assertTrue("LOL" in self.test_case())
		self.assertFalse("KEK" in self.test_case())
		self.assertFalse([ "KEK" ] in self.test_case())




	def test_magical_delitem(self):

		self.test_case("LOL","KEK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case().keys(), [ "LOL" ])
		self.assertEqual(self.test_case().keysof("KEK"), [ "LOL" ])
		self.assertEqual(self.test_case().keysof("CHEBUREK"), [])

		with self.assertLogs("ItemsTest", 10) as case_loggy:

			del self.test_case()["LOL"]
			self.assertEqual(len(self.test_case.real_shelf), 0)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,0)
			del self.test_case()["KEK"]
			del self.test_case()[["KEK"]]


		self.assertEqual(self.test_case().keys(), [])
		self.assertEqual(self.test_case().keysof("KEK"), [])
		self.assertEqual(self.test_case().keysof("CHEBUREK"), [])
		self.assertIn("DEBUG:ItemsTest:Removed \"LOL\", \"KEK\" pair", case_loggy.output)
		self.assertIn(

			f"DEBUG:ItemsTest:Key \"KEK\" not in {self.test_case.magical_shelf} KeyChest",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:ItemsTest:Impossible key \"[\'KEK\']\" for {self.test_case.magical_shelf} KeyChest",
			case_loggy.output
		)








	def test_magical_setitem_silent(self):

		for N,item in enumerate(( "LOL", 1, .1, 0, ( "LOL", )),1):
			with self.subTest(item=item, N=N):

				self.test_case(item, "KEK", silent=True)
				self.assertEqual(len(self.test_case.real_shelf), 0)
				self.assertEqual(len(self.test_case.magical_shelf), N)
				self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
				self.assertEqual(len(self.test_case.magical_shelf._locker_), N)
				self.assertEqual(self.test_case.diff,-N)
				self.assertFalse(self.test_case.modified)




	def test_magical_invalid_setitem_silent(self):

		for item in [ "LOL","KEK" ], { "LOL": "KEK" }, { "LOL","KEK" }:
			with self.subTest(item=item):

				self.test_case(item, "CHEBUREK", silent=True)
				self.assertEqual(len(self.test_case.real_shelf), 0)
				self.assertEqual(len(self.test_case.magical_shelf), 0)
				self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
				self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
				self.assertEqual(self.test_case.diff,0)
				self.assertFalse(self.test_case.modified)




	def test_magical_setitem_edge_case_silent(self):

		self.test_case(None,None, silent=True)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)

		self.test_case(None,None, silent=True)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)

		self.test_case(None,"LOL-KEK-CHEBUREK", silent=True)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)




	def test_magical_replace_item_silent(self):

		self.test_case("LOL","KEK", silent=True)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)

		self.test_case("LOL","CHEBUREK", silent=True)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)

		self.test_case("LOL", "CHEBUREK", silent=True)
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertEqual(self.test_case.diff,-1)
		self.assertFalse(self.test_case.modified)








	def test_real_diff_larger(self):

		self.test_case["LOL"] = "KEK"
		self.test_case.real_shelf("CHE", "BUREK")
		self.assertEqual(len(self.test_case.real_shelf), 2)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case.diff, 1)
		self.assertEqual(self.test_case.real_diff,{ "CHE" })
		self.assertEqual(self.test_case.magical_diff, set())




	def test_magical_diff_larger(self):

		self.test_case["LOL"] = "KEK"
		self.test_case.magical_shelf("CHE", "BUREK")
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 2)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 2)
		self.assertTrue(self.test_case.modified)
		self.assertEqual(self.test_case.diff, -1)
		self.assertEqual(self.test_case.magical_diff,{ "CHE" })
		self.assertEqual(self.test_case.real_diff, set())




	def test_empties_diff(self):

		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertFalse(self.test_case.modified)
		self.assertEqual(self.test_case.diff, 0)
		self.assertEqual(self.test_case.real_diff, set())
		self.assertEqual(self.test_case.magical_diff, set())




	def test_empty_real_diff(self):

		self.test_case.real_shelf("LOL", "KEK")
		self.assertEqual(len(self.test_case.real_shelf), 1)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 1)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertFalse(self.test_case.modified)
		self.assertEqual(self.test_case.diff, 1)
		self.assertEqual(self.test_case.real_diff, { "LOL" })
		self.assertEqual(self.test_case.magical_diff, set())




	def test_empty_magical_diff(self):

		self.test_case.magical_shelf("LOL", "KEK")
		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 1)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 1)
		self.assertFalse(self.test_case.modified)
		self.assertEqual(self.test_case.diff, -1)
		self.assertEqual(self.test_case.magical_diff, { "LOL" })
		self.assertEqual(self.test_case.real_diff, set())
















class GrabbingTest(IrmaTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:

			if	os.path.isfile(cls.GRABBING_IN_INIT):			os.remove(cls.GRABBING_IN_INIT)
			if	os.path.isfile(cls.GRABBING_HANDLER):			os.remove(cls.GRABBING_HANDLER)

			if	os.path.isfile(cls.GRABBING_LOCKER):			os.remove(cls.GRABBING_LOCKER)
			if	os.path.isfile(cls.GRABBING_NO_LOCKER):			os.remove(cls.GRABBING_NO_LOCKER)

			if	os.path.isfile(cls.GRABBING_LOCKER + ".db"):	os.remove(cls.GRABBING_LOCKER + ".db")
			if	os.path.isfile(cls.GRABBING_NO_LOCKER + ".db"):	os.remove(cls.GRABBING_NO_LOCKER + ".db")

			if	os.path.isfile(cls.GRABBING_LOCKER + ".dat"):	os.remove(cls.GRABBING_LOCKER + ".dat")
			if	os.path.isfile(cls.GRABBING_LOCKER + ".bak"):	os.remove(cls.GRABBING_LOCKER + ".bak")
			if	os.path.isfile(cls.GRABBING_LOCKER + ".dir"):	os.remove(cls.GRABBING_LOCKER + ".dir")

			if	os.path.isfile(cls.GRABBING_NO_LOCKER + ".dat"):
				os.remove(cls.GRABBING_NO_LOCKER + ".dat")

			if	os.path.isfile(cls.GRABBING_NO_LOCKER + ".bak"):
				os.remove(cls.GRABBING_NO_LOCKER + ".bak")

			if	os.path.isfile(cls.GRABBING_NO_LOCKER + ".dir"):
				os.remove(cls.GRABBING_NO_LOCKER + ".dir")


	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.GRABBING_HANDLER)

		class Preparation(LibraryShelf):	pass

		preparation = Preparation()
		preparation["OOH"] = "EEH"
		preparation["OOH"] = "AAH-AAH"
		preparation["DING"] = "DANG"
		preparation["WOLO-WOLO"] = "BANG-BANG"
		preparation.produce(cls.GRABBING_NO_LOCKER, rewrite=True)
		preparation.produce(cls.GRABBING_LOCKER, rewrite=True, locker_mode=True)


		class GrabbingTestCase(LibraryShelf):
			class loggy(LibraryContrib):

				init_name	= "GrabbingTest"
				init_level	= 10
				handler		= cls.GRABBING_HANDLER

		cls.test_case = GrabbingTestCase()

	def setUp(self):	self.test_case.unload(magical=True)
	def test_grabbing_no_argument(self):

		with self.assertLogs("GrabbingTest", 10) as case_loggy:	self.test_case.grab()
		self.assertIn("DEBUG:GrabbingTest:Shelf to grab not provided", case_loggy.output)


		self.assertTrue(

			os.path.isfile(self.GRABBING_LOCKER)
			or
			os.path.isfile(self.GRABBING_LOCKER + ".db")
			or
			os.path.isfile(self.GRABBING_LOCKER + ".dat")
		)
		with self.assertLogs("GrabbingTest", 10) as case_loggy:

			self.test_case.grabbing = self.GRABBING_LOCKER
			self.test_case.grab()

		self.assertIn(
			f"DEBUG:GrabbingTest:Shelf \"{self.GRABBING_LOCKER}\" successfully grabbed", case_loggy.output
		)
		self.assertFalse(self.test_case.modified)




	def test_grabbing_invalid_arguments(self):
		self.test_case.grabbing = None

		for grabbable in False,"",[],{},str(),list(),set(),dict(),0,.0 :
			with self.subTest(grabbable=grabbable):

				with self.assertLogs("GrabbingTest", 10) as case_loggy : self.test_case.grab(grabbable)
				self.assertIn(f"DEBUG:GrabbingTest:Shelf to grab not provided", case_loggy.output)




	def test_grabbing_invalid_argument(self):

		for grabbable in (

			True,
			100500,
			420.69,
			unittest,
			...,
			[ "/SomeTest.Shelf" ],
			( "/SomeTest.Shelf", ),
			{ "/SomeTest.Shelf" },
			{ "path": "/SomeTest.Shelf" },
		):
			with self.subTest(grabbable=grabbable):

				with self.assertLogs("GrabbingTest", 10) as case_loggy : self.test_case.grab(grabbable)
				self.assertIn(

					f"WARNING:GrabbingTest:Invalid grabbing path type \"{type(grabbable)}\"",
					case_loggy.output
				)




	def test_grabbing_in_init(self):
		class GrabbingInInnit(LibraryShelf):

			grabbing = self.GRABBING_LOCKER
			class loggy(LibraryContrib):

				init_name	= "GrabbingInInnit"
				init_level	= 10
				handler		= self.GRABBING_IN_INIT


		self.make_loggy_file(self.GRABBING_IN_INIT)
		with self.assertLogs("GrabbingInInnit", 10) as grabbing_loggy:	current_case = GrabbingInInnit()

		self.assertIn("DEBUG:GrabbingInInnit:Obtained locker of length 3", grabbing_loggy.output)
		self.assertIn(

			f"DEBUG:GrabbingInInnit:Shelf \"{self.GRABBING_LOCKER}\" successfully grabbed",
			grabbing_loggy.output
		)
		self.assertEqual(len(current_case.real_shelf), 3)
		self.assertEqual(len(current_case.magical_shelf), 0)
		self.assertEqual(len(current_case.real_shelf._locker_), 3)
		self.assertEqual(len(current_case.magical_shelf._locker_), 0)
		self.assertEqual(current_case.diff,3)
		self.assertFalse(current_case.modified)
		# No "_locker_" in "keys" as it supposed to be
		self.assertEqual(current_case.keys(), [ "OOH", "DING", "WOLO-WOLO" ])
		self.assertEqual(current_case.keysof("EEH"), [])
		self.assertEqual(current_case.keysof("AAH-AAH"), [ "OOH" ])
		self.assertEqual(current_case.keysof("DANG"), [ "DING" ])
		self.assertEqual(current_case.keysof("BANG-BANG"), [ "WOLO-WOLO" ])
		current_case.loggy.handler.close()




	def test_grabbing_with_locker(self):

		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertEqual(self.test_case.diff,0)

		with self.assertLogs("GrabbingTest", 10) as case_loggy:
			self.assertEqual(
				self.test_case.grab(

					self.GRABBING_LOCKER,
					from_locker=True
				),	self.GRABBING_LOCKER
			)

		self.assertIn("DEBUG:GrabbingTest:Obtained locker of length 3", case_loggy.output)
		self.assertIn(
			f"DEBUG:GrabbingTest:Shelf \"{self.GRABBING_LOCKER}\" successfully grabbed", case_loggy.output
		)
		self.assertEqual(self.test_case.real_shelf._locker_, [ "OOH", "DING", "WOLO-WOLO" ])
		self.assertEqual(len(self.test_case.real_shelf), 3)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 3)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertEqual(self.test_case.diff,3)
		# No "_locker_" in "keys" as it supposed to be
		self.assertEqual(self.test_case.keys(), [ "OOH", "DING", "WOLO-WOLO" ])
		self.assertEqual(self.test_case.keysof("EEH"), [])
		self.assertEqual(self.test_case.keysof("AAH-AAH"), [ "OOH" ])
		self.assertEqual(self.test_case.keysof("DANG"), [ "DING" ])
		self.assertEqual(self.test_case.keysof("BANG-BANG"), [ "WOLO-WOLO" ])




	def test_grabbing_with_skipped_locker(self):

		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertEqual(self.test_case.diff,0)

		with self.assertLogs("GrabbingTest", 10) as case_loggy:
			self.assertEqual(
				self.test_case.grab(

					self.GRABBING_LOCKER,
					from_locker=False
				),	self.GRABBING_LOCKER
			)

		self.assertIn("DEBUG:GrabbingTest:Locker skipped, order not granted", case_loggy.output)
		self.assertIn(
			f"DEBUG:GrabbingTest:Shelf \"{self.GRABBING_LOCKER}\" successfully grabbed", case_loggy.output
		)
		self.assertEqual(len(self.test_case.real_shelf), 3)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 3)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertEqual(self.test_case.diff,3)




	def test_grabbing_with_no_locker(self):

		self.assertEqual(len(self.test_case.real_shelf), 0)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertEqual(self.test_case.diff,0)

		with self.assertLogs("GrabbingTest", 10) as case_loggy:
			self.assertEqual(
				self.test_case.grab(

					self.GRABBING_NO_LOCKER,
					from_locker=True
				),	self.GRABBING_NO_LOCKER
			)

		self.assertIn("DEBUG:GrabbingTest:Locker not obtained, order not granted", case_loggy.output)
		self.assertIn(

			f"DEBUG:GrabbingTest:Shelf \"{self.GRABBING_NO_LOCKER}\" successfully grabbed",
			case_loggy.output
		)
		self.assertEqual(len(self.test_case.real_shelf), 3)
		self.assertEqual(len(self.test_case.magical_shelf), 0)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 3)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
		self.assertEqual(self.test_case.diff,3)
















class ProducingTest(IrmaTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:

			if	os.path.isfile(cls.PRODUCING_HANDLER):		os.remove(cls.PRODUCING_HANDLER)
			if	os.path.isfile(cls.PRODUCING_PRODUCABLE):	os.remove(cls.PRODUCING_PRODUCABLE)

			if	os.path.isfile(cls.PRODUCING_PRODUCABLE + ".db"):
				os.remove(cls.PRODUCING_PRODUCABLE + ".db")

			if	os.path.isfile(cls.PRODUCING_PRODUCABLE + ".dat"):
				os.remove(cls.PRODUCING_PRODUCABLE + ".dat")

			if	os.path.isfile(cls.PRODUCING_PRODUCABLE + ".bak"):
				os.remove(cls.PRODUCING_PRODUCABLE + ".bak")

			if	os.path.isfile(cls.PRODUCING_PRODUCABLE + ".dir"):
				os.remove(cls.PRODUCING_PRODUCABLE + ".dir")

	@classmethod
	def setUpClass(cls):

		class ProducingTestCase(LibraryShelf):
			class loggy(LibraryContrib):

				init_name	= "ProducingTest"
				init_level	= 10
				handler		= cls.PRODUCING_HANDLER


		cls.make_loggy_file(cls, cls.PRODUCING_HANDLER)
		cls.test_case	= ProducingTestCase()

	def setUp(self):

		if	os.path.isfile(self.PRODUCING_PRODUCABLE):	os.remove(self.PRODUCING_PRODUCABLE)
		self.test_case.unload(magical=True)
		self.test_case["LOL"] = "KEK"
		self.test_case["CHE"] = "BUREK"


	def test_producing_real(self):

		self.assertEqual(len(self.test_case.real_shelf), 2)
		self.assertEqual(len(self.test_case.magical_shelf), 2)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 2)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)

		with self.assertLogs("ProducingTest", 10) as case_loggy:

			self.test_case.produce(self.PRODUCING_PRODUCABLE, rewrite=True)
			self.test_case.unload(magical=True)

			self.assertEqual(len(self.test_case.real_shelf), 0)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,0)

			self.test_case.grab(self.PRODUCING_PRODUCABLE)
			self.assertEqual(len(self.test_case.real_shelf), 2)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,2)

			self.test_case.produce(self.PRODUCING_PRODUCABLE)

		self.assertIn(

			f"INFO:ProducingTest:Rewritten Shelf \"{self.PRODUCING_PRODUCABLE}\"",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:ProducingTest:Produced key \"LOL\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Produced key \"CHE\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Shelf size 2 keys", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Locker not obtained, order not granted", case_loggy.output)




	def test_producing_real_with_locker(self):

		self.assertEqual(len(self.test_case.real_shelf), 2)
		self.assertEqual(len(self.test_case.magical_shelf), 2)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 2)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)

		with self.assertLogs("ProducingTest", 10) as case_loggy:

			self.test_case.produce(self.PRODUCING_PRODUCABLE, locker_mode=True)
			self.test_case.unload(magical=True)

			self.assertEqual(len(self.test_case.real_shelf), 0)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,0)

			self.test_case.grab(self.PRODUCING_PRODUCABLE)
			self.assertEqual(len(self.test_case.real_shelf), 2)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,2)

			self.test_case.produce(self.PRODUCING_PRODUCABLE)

		self.assertIn(f"DEBUG:ProducingTest:Produced key \"LOL\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Produced key \"CHE\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Locker of length 2 produced", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Shelf size 3 keys", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Obtained locker of length 2", case_loggy.output)




	def test_producing_magical(self):

		self.assertEqual(len(self.test_case.real_shelf), 2)
		self.assertEqual(len(self.test_case.magical_shelf), 2)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 2)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)

		with self.assertLogs("ProducingTest", 10) as case_loggy:

			self.test_case.produce(self.PRODUCING_PRODUCABLE, rewrite=True, magical=True)
			self.test_case.unload(magical=True)

			self.assertEqual(len(self.test_case.real_shelf), 0)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,0)

			self.test_case.grab(self.PRODUCING_PRODUCABLE)
			self.assertEqual(len(self.test_case.real_shelf), 2)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,2)

			self.test_case.produce(self.PRODUCING_PRODUCABLE)

		self.assertIn(f"DEBUG:ProducingTest:Source is magical Shelf with 2 keys", case_loggy.output)
		self.assertIn(

			f"INFO:ProducingTest:Rewritten Shelf \"{self.PRODUCING_PRODUCABLE}\"",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:ProducingTest:Produced key \"LOL\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Produced key \"CHE\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Shelf size 2 keys", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Locker not obtained, order not granted", case_loggy.output)




	def test_producing_magical_with_locker(self):

		self.assertEqual(len(self.test_case.real_shelf), 2)
		self.assertEqual(len(self.test_case.magical_shelf), 2)
		self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
		self.assertEqual(len(self.test_case.magical_shelf._locker_), 2)
		self.assertEqual(self.test_case.diff,0)
		self.assertTrue(self.test_case.modified)

		with self.assertLogs("ProducingTest", 10) as case_loggy:

			self.test_case.produce(self.PRODUCING_PRODUCABLE, magical=True, locker_mode=True)
			self.test_case.unload(magical=True)

			self.assertEqual(len(self.test_case.real_shelf), 0)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 0)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,0)

			self.test_case.grab(self.PRODUCING_PRODUCABLE)
			self.assertEqual(len(self.test_case.real_shelf), 2)
			self.assertEqual(len(self.test_case.magical_shelf), 0)
			self.assertEqual(len(self.test_case.real_shelf._locker_), 2)
			self.assertEqual(len(self.test_case.magical_shelf._locker_), 0)
			self.assertEqual(self.test_case.diff,2)

			self.test_case.produce(self.PRODUCING_PRODUCABLE)

		self.assertIn(f"DEBUG:ProducingTest:Source is magical Shelf with 2 keys", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Produced key \"LOL\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Produced key \"CHE\"", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Locker of length 2 produced", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Shelf size 3 keys", case_loggy.output)
		self.assertIn(f"DEBUG:ProducingTest:Obtained locker of length 2", case_loggy.output)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pathlib								import Path
from	shutil								import copyfile
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.hagrid				import EasySet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.planting.leafs		import LeafProbe
from	pygwarts.hagrid.bloom.leafs			import Rejuvenation








class RejuvenationCase(EasySet):

	"""
		Custom grow (situative, must raise FileNotFoundError) for Rejuvenation
		Single sprout
		Single bough
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.EASY_HANDLER_6): os.remove(cls.EASY_HANDLER_6)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.EASY_HANDLER_6)
		class Sakura(Tree):

			bough = cls.EASY_SET_BOUGH
			class loggy(LibraryContrib):

				init_name	= "rejuvenations"
				handler		= cls.EASY_HANDLER_6
				init_level	= 10

			class files(Rejuvenation):		pass
			class pseudogrow(LeafProbe):	pass
			class grow(Transmutable):

				def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):
					copyfile(leaf, bough.joinpath(leaf.name))

		cls.notdst1 = os.path.join(cls.EASY_SET_BOUGH, "not.exist")
		cls.notdst2 = os.path.join(cls.EASY_SET_BOUGH, "not.exist1")
		cls.notdst3 = os.path.join(cls.EASY_SET_BOUGH, "not.exist2")
		cls.notdst4 = os.path.join(cls.EASY_SET_BOUGH, "not.exist3")
		cls.notsrc1 = os.path.join(cls.EASY_SET_SPROUT, "not.exist")
		cls.notsrc2 = os.path.join(cls.EASY_SET_SPROUT, "not.exist1")
		cls.notsrc3 = os.path.join(cls.EASY_SET_SPROUT, "not.exist2")
		cls.notsrc4 = os.path.join(cls.EASY_SET_SPROUT, "not.exist3")
		cls.notsrc5 = os.path.join(cls.EASY_SET_SPROUT, "non-existant folder")
		cls.notsrc6 = os.path.join(cls.notsrc5, "not.exist1")
		cls.notsrc7 = os.path.join(cls.notsrc5, "not.exist2")
		cls.notsrc8 = os.path.join(cls.notsrc5, "not.exist3")
		cls.test_case = Sakura()




	def test_initiate_rejuve(self):


		self.assertTrue(len(self.test_case) == 1)
		self.assertTrue(str(self.test_case[0]) == "Sakura.files")
		self.assertTrue(isinstance(self.test_case[0], Rejuvenation))
		self.assertTrue(hasattr(self.test_case.files, "blooming"))
		self.assertTrue(isinstance(self.test_case.files.blooming, dict))
		self.assertEqual(len(self.test_case.files.blooming),0)




	def test_j_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist") ]
			)

		self.assertEqual(len(self.test_case.files.blooming),0)
		self.assertNotIn(str(self.test_case), self.test_case.files.blooming)
		self.assertIn(
			str(

				"WARNING:rejuvenations:Failed to rejuve leaf "
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist')}\" due to FileNotFoundError: "
				f"[Errno 2] No such file or directory: \'{self.notsrc1}\'"
			)	if os.name == "posix" else "%s%s%s'%s'"%(
				"WARNING:rejuvenations:Failed to rejuve leaf ",
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist')}\" due to FileNotFoundError: ",
				f"[Errno 2] No such file or directory: ",
				self.notsrc1.replace("\\", "\\\\")
			),
			case_loggy.output
		)




	def test_k_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[
					Path(self.EASY_SET_SPROUT).joinpath("not.exist1"),
					Path(self.EASY_SET_SPROUT).joinpath("not.exist2"),
					Path(self.EASY_SET_SPROUT).joinpath("not.exist3")
				]
			)

		self.assertEqual(len(self.test_case.files.blooming),0)
		self.assertNotIn(str(self.test_case), self.test_case.files.blooming)
		self.assertIn(
			str(
				"WARNING:rejuvenations:Failed to rejuve leaf "
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist1')}\" due to "
				f"FileNotFoundError: [Errno 2] No such file or directory: \'{self.notsrc2}\'"
			)	if os.name == "posix" else "%s%s%s'%s'"%(
				"WARNING:rejuvenations:Failed to rejuve leaf ",
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist1')}\" due to ",
				f"FileNotFoundError: [Errno 2] No such file or directory: ",
				self.notsrc2.replace("\\", "\\\\")
			),
			case_loggy.output
		)
		self.assertIn(
			str(
				"WARNING:rejuvenations:Failed to rejuve leaf "
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist2')}\" due to "
				f"FileNotFoundError: [Errno 2] No such file or directory: \'{self.notsrc3}\'"
			)	if os.name == "posix" else "%s%s%s'%s'"%(
				"WARNING:rejuvenations:Failed to rejuve leaf ",
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist2')}\" due to ",
				f"FileNotFoundError: [Errno 2] No such file or directory: ",
				self.notsrc3.replace("\\", "\\\\")
			),
			case_loggy.output
		)
		self.assertIn(
			str(
				"WARNING:rejuvenations:Failed to rejuve leaf "
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist3')}\" due to "
				f"FileNotFoundError: [Errno 2] No such file or directory: \'{self.notsrc4}\'"
			)	if os.name == "posix" else "%s%s%s'%s'"%(
				"WARNING:rejuvenations:Failed to rejuve leaf ",
				f"\"{Path(self.EASY_SET_SPROUT).joinpath('not.exist3')}\" due to ",
				f"FileNotFoundError: [Errno 2] No such file or directory: ",
				self.notsrc4.replace("\\", "\\\\")
			),
			case_loggy.output
		)




	def test_l_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[]
			)

		self.assertEqual(len(self.test_case.files.blooming),0)
		self.assertNotIn(str(self.test_case), self.test_case.files.blooming)
		self.assertIn(

			f"DEBUG:rejuvenations:Branch \"{self.EASY_SET_SPROUT}\" has no leafs to rejuve",
			case_loggy.output
		)




	def test_m_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist") ]
			)

		self.assertEqual(len(self.test_case.files.blooming),1)
		self.assertIn("self.test_case", self.test_case.files.blooming)
		self.assertNotIn("grow", self.test_case.files.blooming["self.test_case"])
		self.assertFalse(self.test_case.files.blooming["self.test_case"]["bough"])
		self.assertIn(

			f"ERROR:rejuvenations:Invalid tree self.test_case or no bough to rejuve",
			case_loggy.output
		)




	def test_n_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[
					Path(self.EASY_SET_SPROUT).joinpath("not.exist")
				]
			)

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist1") ]
			)


		self.assertEqual(len(self.test_case.files.blooming),1)
		self.assertIn("self.test_case", self.test_case.files.blooming)
		self.assertNotIn("grow", self.test_case.files.blooming["self.test_case"])
		self.assertFalse(self.test_case.files.blooming["self.test_case"]["bough"])
		self.assertIn(

			f"DEBUG:rejuvenations:Invalid tree self.test_case or no bough to rejuve",
			case_loggy.output
		)
		self.assertEqual(

			case_loggy.output.count(
				f"DEBUG:rejuvenations:Invalid tree self.test_case or no bough to rejuve"
			),2
		)




	def test_o_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist2") ]
			)

			self.assertFalse(self.test_case.files.blooming["self.test_case"]["bough"])
			self.test_case.files.blooming["self.test_case"]["bough"] = True
			self.assertTrue(self.test_case.files.blooming["self.test_case"]["bough"])

			self.test_case[0](

				"self.test_case",
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist3") ]
			)

		self.assertEqual(len(self.test_case.files.blooming),1)
		self.assertIn("self.test_case", self.test_case.files.blooming)
		self.assertNotIn("grow", self.test_case.files.blooming["self.test_case"])
		self.assertFalse(self.test_case.files.blooming["self.test_case"]["bough"])
		self.assertIn(

			f"DEBUG:rejuvenations:Invalid tree self.test_case or no bough to rejuve",
			case_loggy.output
		)
		self.assertEqual(

			case_loggy.output.count(
				f"DEBUG:rejuvenations:Invalid tree self.test_case or no bough to rejuve"
			),1
		)
		self.assertIn(

			f"ERROR:rejuvenations:Invalid tree self.test_case or no bough to rejuve",
			case_loggy.output
		)
		self.assertEqual(

			case_loggy.output.count(
				f"ERROR:rejuvenations:Invalid tree self.test_case or no bough to rejuve"
			),1
		)




	def test_p_rejuve(self):

		self.test_case.grow = 42
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist") ]
			)

		self.assertEqual(len(self.test_case.files.blooming),2)
		self.assertIn(str(self.test_case), self.test_case.files.blooming)
		self.assertFalse(self.test_case.files.blooming[str(self.test_case)]["grow"])
		self.assertNotIn("bough", self.test_case.files.blooming[str(self.test_case)])
		self.assertIn(f"ERROR:rejuvenations:{self.test_case} doesn't implement grow", case_loggy.output)




	def test_q_rejuve(self):

		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist") ]
			)
			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist1") ]
			)

		self.assertEqual(len(self.test_case.files.blooming),2)
		self.assertIn(str(self.test_case), self.test_case.files.blooming)
		self.assertFalse(self.test_case.files.blooming[str(self.test_case)]["grow"])
		self.assertNotIn("bough", self.test_case.files.blooming[str(self.test_case)])
		self.assertIn(f"DEBUG:rejuvenations:{self.test_case} doesn't implement grow", case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:rejuvenations:{self.test_case} doesn't implement grow"),2
		)




	def test_r_rejuve(self):

		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist2") ]
			)

			self.assertFalse(self.test_case.files.blooming[str(self.test_case)]["grow"])
			self.test_case.files.blooming[str(self.test_case)]["grow"] = True
			self.assertTrue(self.test_case.files.blooming[str(self.test_case)]["grow"])

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist3") ]
			)

		self.assertEqual(len(self.test_case.files.blooming),2)
		self.assertIn(str(self.test_case), self.test_case.files.blooming)
		self.assertFalse(self.test_case.files.blooming[str(self.test_case)]["grow"])
		self.assertNotIn("bough", self.test_case.files.blooming[str(self.test_case)])
		self.assertIn(f"DEBUG:rejuvenations:{self.test_case} doesn't implement grow", case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:rejuvenations:{self.test_case} doesn't implement grow"),1
		)
		self.assertIn(f"ERROR:rejuvenations:{self.test_case} doesn't implement grow", case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"ERROR:rejuvenations:{self.test_case} doesn't implement grow"),1
		)




	def test_s_rejuve(self):

		self.test_case.grow = self.test_case.pseudogrow
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[ Path(self.EASY_SET_SPROUT).joinpath("not.exist") ]
			)

		self.assertEqual(len(self.test_case.files.blooming),2)
		self.assertIn(str(self.test_case), self.test_case.files.blooming)
		self.assertTrue(self.test_case.files.blooming[str(self.test_case)]["grow"])
		self.assertTrue(self.test_case.files.blooming[str(self.test_case)]["bough"])
		self.assertIn(f"DEBUG:rejuvenations:Branch \"{self.notsrc1}\" not located", case_loggy.output)




	def test_t_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(self.EASY_SET_SPROUT),
				[],
				[
					Path(self.EASY_SET_SPROUT).joinpath("not.exist1"),
					Path(self.EASY_SET_SPROUT).joinpath("not.exist2"),
					Path(self.EASY_SET_SPROUT).joinpath("not.exist3")
				]
			)

		self.assertEqual(len(self.test_case.files.blooming),2)
		self.assertIn(str(self.test_case), self.test_case.files.blooming)
		self.assertTrue(self.test_case.files.blooming[str(self.test_case)]["grow"])
		self.assertTrue(self.test_case.files.blooming[str(self.test_case)]["bough"])
		self.assertIn(f"DEBUG:rejuvenations:Branch \"{self.notsrc2}\" not located", case_loggy.output)
		self.assertIn(f"DEBUG:rejuvenations:Branch \"{self.notsrc3}\" not located", case_loggy.output)
		self.assertIn(f"DEBUG:rejuvenations:Branch \"{self.notsrc4}\" not located", case_loggy.output)




	def test_u_rejuve(self):
		with self.assertLogs("rejuvenations", 10) as case_loggy:

			self.test_case[0](

				self.test_case,
				self.EASY_SET_SPROUT,
				Path(os.path.join(self.EASY_SET_SPROUT, "non-existant folder")),
				[],
				[
					Path(os.path.join(self.EASY_SET_SPROUT, "non-existant folder")).joinpath("not.exist1"),
					Path(os.path.join(self.EASY_SET_SPROUT, "non-existant folder")).joinpath("not.exist2"),
					Path(os.path.join(self.EASY_SET_SPROUT, "non-existant folder")).joinpath("not.exist3")
				]
			)

		self.assertEqual(len(self.test_case.files.blooming),2)
		self.assertIn(str(self.test_case), self.test_case.files.blooming)
		self.assertTrue(self.test_case.files.blooming[str(self.test_case)]["grow"])
		self.assertTrue(self.test_case.files.blooming[str(self.test_case)]["bough"])
		self.assertIn(f"DEBUG:rejuvenations:Branch \"{self.notsrc6}\" not located", case_loggy.output)
		self.assertIn(f"DEBUG:rejuvenations:Branch \"{self.notsrc7}\" not located", case_loggy.output)
		self.assertIn(f"DEBUG:rejuvenations:Branch \"{self.notsrc8}\" not located", case_loggy.output)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	unittest
from	pygwarts.tests.irma.unit_irma_contrib			import *
from	pygwarts.tests.irma.unit_irma_contrib_intercept	import *
from	pygwarts.tests.irma.unit_irma_access			import *
from	pygwarts.tests.irma.unit_irma_access_volume		import *
from	pygwarts.tests.irma.unit_irma_access_bookmark	import *
from	pygwarts.tests.irma.unit_irma_access_handler	import *
from	pygwarts.tests.irma.unit_irma_access_inducer	import *
from	pygwarts.tests.irma.unit_irma_access_annex		import *
from	pygwarts.tests.irma.unit_irma_access_utils		import *
from	pygwarts.tests.irma.unit_irma_shelf				import *
from	pygwarts.tests.irma.unit_irma_shelf_casing		import *








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	unittest
from	pygwarts.tests.magical.unit_magical_chest						import *
from	pygwarts.tests.magical.unit_magical_spells						import *
from	pygwarts.tests.magical.unit_magical_key_chest					import *
from	pygwarts.tests.magical.unit_magical_time_turner					import *
from	pygwarts.tests.magical.unit_magical_time_turner_utils			import *
from	pygwarts.tests.magical.unit_magical_time_turner_timers			import *
from	pygwarts.tests.magical.unit_magical_phstone						import *
from	pygwarts.tests.magical.unit_magical_transmutations				import *
from	pygwarts.tests.magical.unit_magical_controlled_transmutations	import *








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
from	pathlib			import Path
from	pygwarts.tests	import PygwartsTestCase








class HedwigTestCase(PygwartsTestCase):

	"""
		pygwarts.hedwig tests super class
	"""

	HEDWIG_ROOT					= Path(PygwartsTestCase._PygwartsTestCase__CWD) /"hedwig"
	HEDWIG_MAIL_UTILS_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"hedwig" /"mu.loggy")
	HEDWIG_MAIL_LETTER_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"hedwig" /"ml.loggy")
	HEDWIG_MAIL_LETTER_AHANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"hedwig" /"mla.loggy")
	HEDWIG_MAIL_BUILDER_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"hedwig" /"mb.loggy")
	HEDWIG_MAIL_CLIENT_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"hedwig" /"mc.loggy")
	HEDWIG_MAIL_SMTP_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"hedwig" /"ms.loggy")
	VALID_EMAILS = (

		"email@example.com",
		"firstname.lastname@example.com",
		"EMAIL@subdomain.example.com",
		"firstname+lastname@example.com",
		"email@123.123.123.123",
		"1234567890@example.com",
		"email@example-one.com",
		"_______@example.com",
		"email@example.name",
		"email@example.museum",
		"email@example.co.jp",
		"firstname-lastname@example.com",
		"email@example.web",
		"email@111.222.333.44444",
	)
	INVALID_EMAILS = (

		"",
		"plainaddress",
		"#@%^%#$@#$@#.com",
		"@example.com",
		"Joe Smith <email@example.com>",
		"email.example.com",
		"email@example@example.com",
		".email@example.com",
		"email.@example.com",
		"email..email@example.com",
		"email@[123.123.123.123]",
		"email@example.com (Joe Smith)",
		"email@example",
		"email@-example.com",
		"email@example..com",
		"Abc..123@example.com",
		" email@example.com",
		"em;ail@example.com",
		"email@ex;ample.com",
	)










from os													import sep
from pathlib											import Path
from typing												import Tuple
from typing												import Callable
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import Transmutation
from pygwarts.magical.philosophers_stone.transmutations	import ControlledTransmutation
from pygwarts.magical.spells							import geminio
from pygwarts.hagrid.thrivables							import Tree








class GrowingPeel(Transmutation):

	"""
		hagrid.planting utility decorator class, that helps in obtaining corresponding destination "sprigs"
		(folders/files) as Path objects. As Transmutation class, in mutable chain takes decorated "planting"
		class and extends it by declaring meta __call__ to invoke decorated __call__. Meta __call__ designed
		to, and decorated __call__ must, accept arguments according to "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			twig	- Path object that represents current operation source folder;
			bough	- Path object that represents current operation destination parent folder.
		Meta __call__ will take current source "sprig", which must be Path object "branch", obtain relative
		to it's source root subpath and concatenate it to destination root "bough". This final concatenation
		will represent corresponding destination "sprig".
		Every logging will refer to the "origin" Tree object.
	"""

	def _mutable_chain_injection(self, mutable_layer :Transmutable) -> Transmutable :
		class Peel(geminio(mutable_layer)):


			def __call__(self, origin :Tree, sprout :str, branch :Path, sprig :Path, bough :Path):


				if	isinstance(branch, Path):
					if	isinstance(bough, Path):


						grown_bough = bough.joinpath(branch.relative_to(sprout))
						origin.loggy.debug(f"Peeled growing bough \"{grown_bough}\"")
						super().__call__(origin, sprout, branch, sprig, grown_bough)
					else:
						self.loggy.debug(f"Bough \"{bough}\" is not a Path object")
				else:	self.loggy.debug(f"Branch \"{branch}\" is not a Path object")
		return	Peel








class ThrivingPeel(ControlledTransmutation):

	"""
		hagrid.planting utility decorator class, that helps in obtaining corresponding destination "sprigs"
		(folders/files) as Path objects. As ControlledTransmutation class, accepts all positional arguments
		as "thrivings", which must be strings that represent intermediate folders to be integrated in final
		destination "sprig" path. Also accepts boolean flag "to_peak" which regulates whether collected
		"thrivings" must be added to the very end of destination "sprig" path or in the middle (defaulted
		to True - to the top). In mutable chain acts as a mutation - takes decorated "planting" class and
		extends it by declaring meta __call__ to invoke decorated __call__. Meta __call__ designed to, and
		decorated __call__ must, accept arguments according to "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			twig	- Path object that represents current operation source folder;
			bough	- Path object that represents current operation destination parent folder.
		Meta __call__ will take current source "sprig", which must be Path object "branch", obtain relative
		to it's source root subpath, merge it with "thrivings" subfolders and concatenate it to destination
		root "bough". This final concatenation will represent corresponding destination "sprig".
		Every logging will refer to the "origin" Tree object.
		ThrivingPeel with "to_peak" set to True has some sort of side effect when used with "Germination",
		which causes every folder to be created with preceeding "thrivings". For files this behaviour is
		desired, but for folders this is might be an unpredictable source of bags in execution scripts.
		As "Germination" designed for full replication of source tree, it is not recommended to decorate
		it with ThrivingPeel.
	"""

	def __init__(self, *thrivings :Tuple[str,], to_peak=True):


		self.to_peak = to_peak
		self.thrivings = list()


		# The only condition to comply for thriving target directory, is that such "thriving"
		# argument is a string. To avoid something like PermissionError when "thriving" provided
		# with leading fs-separation symbol, it's stripped out.
		for thriving in thrivings:

			assert isinstance(thriving, str), f"Thriving must be with strings, not \"{type(thriving)}\""
			self.thrivings.append(thriving.lstrip(sep))


	def _mutable_chain_injection(self, mutable_layer :Transmutable) -> Transmutable :


		thrivings	= self.thrivings
		to_peak		= self.to_peak


		class Peel(geminio(mutable_layer)):
			def __call__(self, origin :Tree, sprout :str, branch :str, sprig :str, bough :str):


				if	isinstance(branch, Path):
					if	isinstance(bough, Path):
						if	to_peak:


							thrived_bough = bough.joinpath(branch.relative_to(sprout), *thrivings)
						else:
							thrived_bough = bough.joinpath(*thrivings, branch.relative_to(sprout))


						origin.loggy.debug(f"Peeled thriving bough \"{thrived_bough}\"")
						super().__call__(origin, sprout, branch, sprig, thrived_bough)
					else:
						self.loggy.debug(f"Bough \"{bough}\" is not a Path object")
				else:	self.loggy.debug(f"Branch \"{branch}\" is not a Path object")
		return	Peel










from pathlib											import Path
from typing												import Callable
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import ControlledTransmutation
from pygwarts.magical.spells							import geminio
from pygwarts.hagrid.thrivables							import Tree
from pygwarts.irma.shelve								import LibraryShelf








class DraftPeek(ControlledTransmutation):

	"""
		hagrid.planting utility decorator class, that serves as "planting regulator". Allows Integration
		of "sprigs" (folders/files) modification time check in to "planting" process, so "planting" invocation
		might rely on their comparison. As ControlledTransmutation class, accepts following arguments when
		initiated as a "planting" object decorator:
			renew		- by default only sprigs with old rings will be handled (that means "renew" flag
						if True restricts planting of source "sprig" if corresponding destination "sprig"
						doesn't exist);
			picky		- use floats representation of rings for pickiness (precision) or cast them to int
						(this solves situations, when cloning sprig mtime impossible and copied sprig will
						have slightly different mtime every time, so casting mtime to int must mitigate that
						effect). Defaulted to False (casting to int).
			cache		- allows (by default) caching mtimes of branches that currently walked to reuse it in
						multi-target situations for breadth-first "Flourish" algorithm.
			comparator	- callable that takes new_ring and old_ring as arguments in that only order and return
						boolean to make a decision of handing further.
		In mutable chain acts as a mutation - takes decorated "planting" class and extends it by declaring
		meta __call__ to invoke decorated __call__. Meta __call__ will consider source "sprig" modification
		time, which might be checked directly or from cache, and compare by "comparator" callable with
		corresponding destination "sprig" modification time, which always will be taken directly, or
		defaulted to zero if corresponding destination "sprig" doesn't exist. By "comparator" callable return
		value, decorated __call__ invocation will be considered. Meta __call__ designed to, and decorated
		__call__ must, accept arguments according to "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			twig	- Path object that represents current operation source folder;
			bough	- Path object that represents current operation destination parent folder.
		Every logging will refer to the "origin" Tree object.
	"""

	def __init__(
					self,
					*,
					renew		:bool										=True,
					picky		:bool										=False,
					cache		:bool										=True,
					comparator	:Callable[[ int|float, int|float ], bool]	=None,
				):


		self.renew	= renew
		self.picky	= picky
		self.cache	= cache


		if	comparator is None : self.comparator = lambda N,O : O <N
		else:
			try:

				if	callable(comparator) and isinstance(comparator(.2,.1), bool):

					self.comparator = comparator
					return
			except:	pass
			raise	TypeError("Draft peek comparator must be Callable[[ int|float, int|float ], bool]")


	def _mutable_chain_injection(self, mutable_layer :Transmutable) -> Transmutable :


		comparator	= self.comparator
		renew		= self.renew
		picky		= self.picky
		cache		= self.cache


		# Temporary location (buffer) for a current branch that is walked, to store all it's peeks,
		# that might be useful for other Trees, cause every Tree's Bloom make a peek for a single sprig
		# every time. If somehow the new ring for some sprig will be altered during all Tree's processing
		# sprigs, it will not affect current runtime cause mtime will be cached. Basically it will not affect
		# current runtime anyway. The idea is that the very first mtime check is based and all further
		# dispatching goes along such obtain.
		walk = dict()


		class Peek(geminio(mutable_layer)):
			def __call__(self, origin :Tree, sprout :str, branch :Path, sprig :Path, bough :Path):


				if	isinstance(sprig, Path) and sprig.is_file() and not sprig.is_symlink():
					old_peek = bough.joinpath(sprig.name)


					if	cache:
						if	(current_branch := walk.get(branch)) is not None:
							if	isinstance(current_branch, dict):
								if	(cached_peek := current_branch.get(str(sprig))) is not None:
									new_ring = cached_peek
					try:

						# As "new_ring" might be taken from cache in not "picky" condition (as int)
						# the following attempt only works in opposite case, when picky cached
						# it turning to integer.
						new_ring = new_ring if picky else int(new_ring)
						origin.loggy.debug(f"Cached draft peek \"{sprig}\" new ring: {new_ring}")


					except	NameError:

						# Modification time is known to be allowed to be peeked for any file in fs,
						# no matter owning and permission bits. However, in case item to peek suddenly
						# disappears or something, wrapping operation in try-except sounds very reasonable.
						# But, such wrapping does nothing with situation, cause no peek means no handing
						# further, hence there is nothing such wrapping can do about it, except raising
						# another Exception. So, it assumed if so incredible situation will occur,
						# some dispatcher will catch it and log.
						new_ring = sprig.stat().st_mtime if picky else int(sprig.stat().st_mtime)
						origin.loggy.debug(f"New draft peek \"{sprig}\" new ring: {new_ring}")


					if	cache:


						try:	walk[branch][str(sprig)] = new_ring
						except	KeyError:

							# New branch is walked so clearing "walk" buffer to save memory
							walk.clear()
							walk[branch] = { str(sprig): new_ring }




					try:	old_ring = old_peek.stat().st_mtime
					except	FileNotFoundError:


						if	not renew:

							old_ring = 0.
						else:
							origin.loggy.debug("Flourish stopped cause draft peek only for sprigs renew")
							return


					old_ring = old_ring if picky else int(old_ring)
					origin.loggy.debug(f"Draft peek for \"{old_peek}\" old ring: {old_ring}")


					if	comparator(new_ring, old_ring):

						super().__call__(origin, sprout, branch, sprig, bough)
					else:
						origin.loggy.debug("Flourish stopped by draft peek comparator")
				else:	origin.loggy.debug(f"Not located draft peek \"{sprig}\"")
		return	Peek








class BlindPeek(ControlledTransmutation):

	"""
		hagrid.planting utility decorator class, that serves as "planting regulator". Allows Integration
		of "sprigs" (folders/files) modification time check in to "planting" process, so "planting" invocation
		might rely on their comparison. As ControlledTransmutation class, accepts following arguments wneh
		initiated as a "planting" object decorator:
			link		- the string to find escalatable LibraryShelf object;
			renew		- by default only sprigs with old rings will be handled (that means "renew" flag
						if True restricts planting of source "sprig" if corresponding destination "sprig"
						doesn't exist);
			picky		- use floats representation of rings for pickiness (precision) or cast them to int
						(this solves situations, when cloning sprig mtime impossible and copied sprig will
						have slightly different mtime every time, so casting mtime to int must mitigate that
						effect). Defaulted to False (casting to int).
			cache		- allows (by default) caching mtimes of branches that currently walked to reuse it in
						multi-target situations for breadth-first "Flourish" algorithm.
			comparator	- callable that takes new_ring and old_ring as arguments in that only order and return
						boolean to make a decision of handing furhter.
		In mutable chain acts as a mutation - takes decorated "planting" class and extends it by declaring
		meta __call__ to invoke decorated __call__. Meta __call__ will consider source "sprig" modification
		time, which might be checked directly or from cache, and compare by "comparator" callable with
		corresponding destination "sprig" modification time, which always will be taken either from
		LibraryShelf object, refered by "link" argument, or defaulted to zero if corresponding destination
		"sprig" doesn't exist, but never directly from corresponding destination "sprig". Attention must be
		payed, that BlindPeek acts really "blind". The name infered from the logic, that "peek" occurs from
		LibraryShelf storage only, despite whether destination "sprig" exists. That might create some bad flow
		when BlindPeek used in the situations, where it is not supposed to. If source "sprig" peek and
		destination record peek comparison is not satisfied, which means that destination record does exist
		in Shelf, but destination "sprig" actually doesn't exist, decorated __call__ will not be invoked by
		comparator. The nearly same situation might occur with cached peeks for source. Situations described
		are not only edge cases but the logic behind the "BlindPeek" algorithm, as it means blind to
		destination "sprigs", operating only with Shelf records. By "comparator" callable return value,
		decorated __call__ invocation will be considered. Meta __call__ designed to, and decorated __call__
		must, accept arguments according to "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			twig	- Path object that represents current operation source folder;
			bough	- Path object that represents current operation destination parent folder.
		Every logging will refer to the "origin" Tree object.
	"""

	def __init__(
					self,
					link		:str,
					*,
					renew		:bool										=True,
					picky		:bool										=False,
					cache		:bool										=True,
					comparator	:Callable[[ int|float, int|float ], bool]	=None,
				):


		self.link	= link
		self.renew	= renew
		self.picky	= picky
		self.cache	= cache


		if	comparator is None : self.comparator = lambda N,O : O <N
		else:

			try:

				if	callable(comparator) and isinstance(comparator(.2,.1), bool):

					self.comparator = comparator
					return
			except:	pass
			raise	TypeError("Blind peek comparator must be Callable[[ int|float, int|float ], bool]")


	def _mutable_chain_injection(self, mutable_layer :Transmutable) -> Transmutable :


		comparator	= self.comparator
		link		= self.link
		renew		= self.renew
		picky		= self.picky
		cache		= self.cache


		# Temporary location (buffer) for a current branch that is walked, to store all it's peeks,
		# that might be useful for other Trees, cause every Tree's Bloom make a peek for a single sprig
		# every time. If somehow the new ring for some sprig will be altered during all Tree's processing
		# sprigs, it will not affect current runtime cause mtime will be cached. Basically it will not affect
		# current runtime anyway. The idea is that the very first mtime check is based and all further
		# dispatching goes along such obtain.
		walk = dict()


		class Peek(geminio(mutable_layer)):
			def __call__(self, origin :Tree, sprout :str, branch :Path, sprig :Path, bough :Path):


				if	isinstance(sprig, Path) and sprig.is_file() and not sprig.is_symlink():
					new_peekable = True


					if	cache:
						if	(current_branch := walk.get(branch)) is not None:
							if	isinstance(current_branch, dict):
								if	(cached_peek := current_branch.get(str(sprig))) is not None:
									new_ring, new_peekable = cached_peek, False
					try:

						new_ring = new_ring if picky else int(new_ring)
						origin.loggy.debug(f"Cached blind peek \"{sprig}\" new ring: {new_ring}")


					except	NameError:

						# Modification time is known to be allowed to be peeked for any file in fs,
						# no matter owning and permission bits. However, in case item to peek suddenly
						# disappears or something, wrapping operation in try-except sounds very reasonable.
						# But, such wrapping does nothing with situation, cause no peek means no handing
						# further, hence there is nothing such wrapping can do about it, except raising
						# another Exception. So, it assumed if so incredible situation will occur,
						# some dispatcher will catch it and log.
						new_ring = sprig.stat().st_mtime if picky else int(sprig.stat().st_mtime)
						origin.loggy.debug(f"New blind peek \"{sprig}\" new ring: {new_ring}")


					if	cache:
						try:	walk[branch][str(sprig)] = new_ring
						except	KeyError:

							# New branch is walked co clearing "walk" buffer to save memory
							walk.clear()
							walk[branch] = { str(sprig): new_ring }




					if	isinstance((seeds := getattr(self, link, None)), LibraryShelf):


						if		(old_ring := seeds[str(sprig)]) is not None: pass
						elif	(not renew): old_ring = 0.
						else:
							origin.loggy.debug("Flourish stopped cause blind peek only for sprigs renew")
							return


						old_ring = old_ring if picky else int(old_ring)
						origin.loggy.debug(f"Blind peek for \"{sprig}\" old ring: {old_ring}")


						if	comparator(new_ring, old_ring):


							# Already cached peeks are stored in temporary dictionary. That allows
							# the flag "new_peekable" to regulate seeding (shelving) by the way it is
							# trigger the "modified" flag or not. That means if current leaf is peeked
							# the very-very first time, it will be seeded. The comparator decides the
							# way it will be seeded, like the negative comparator result leads to
							# silent seeding, because there is no need to reshelve already shelved
							# values. This relies on LibraryShelf logic, when it is called it uses
							# the "magical_shelf" to store key-pairs, and silent mode do the same,
							# but doesn't modify "modified" flag. So if it appears to "modified"
							# flag becomes True on the way, all previous values already in "magical_shelf".
							if	new_peekable : seeds(str(sprig), new_ring)
							super().__call__(origin, sprout, branch, sprig, bough)
						else:
							if	new_peekable : seeds(str(sprig), new_ring, silent=True)
							origin.loggy.debug("Flourish stopped by blind peek comparator")
					else:	origin.loggy.debug("Flourish failed to find \"seeds\" for blind peek")
				else:		origin.loggy.debug(f"Not located blind peek \"{sprig}\"")
		return	Peek


















def is_valid_port(number :int | float | str) -> bool | None :

	"""
		Helper function that verifies "number" is a valid port number by string-casting to int.
		Returns True if it is 0-65535 range integer, or False otherwise. Returns None for the rest.
	"""

	try:	return 0 <= int(str(number)) <= 65535 if isinstance(number, int) else None
	except:	return




def validate_port(number :int | float | str) -> int | None :

	"""
		Helper function that takes "number" argument and triple casts it to int. If obtained int is in
		range 0-65535, returns that int. Returns None in any other cases.
	"""

	try:	port = int(float(str(number)))
	except:	return
	else:	return port if -1 <port <65536 else None







